/***************************************************************************** 
 *
 * xremote - a utility for grabbing the mouse and keyboard of another screen.
 * 
 * Copyright (C) 2000 Jan Ekholm
 *	
 * This program is free software; you can redistribute it and/or modify it  
 * under the terms of the GNU General Public License as published by the  
 * Free Software Foundation; either version 2 of the License, or (at your 
 * option) any later version.
 *	
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License 
 * for more details.
 *	
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *	
 * Email: chakie@infa.abo.fi
 *	
 * For more information about the author or xremote see the homepage: 
 * http://infa.abo.fi/~chakie/xremote/
 *
 ****************************************************************************/

/***************************************************************************** 
 * Do we have config.h?
 ****************************************************************************/
#ifdef HAVE_CONFIG
#include "config.h"
#endif


/***************************************************************************** 
 * Too lazy to use the std:: everywhere
 ****************************************************************************/
using namespace std;

/***************************************************************************** 
 * What iostream do we have?
 ****************************************************************************/
#include <string>
#include <iostream>
#include <fstream>
#include <cstring>      // strcmp

//  #ifdef HAVE_IOSTREAM
//  #else
//  #include <string>
//  #include <iostream.h>
//  #include <fstream.h>
//  #endif

/***************************************************************************** 
 * Includes
 ****************************************************************************/
#include <stdio.h>		
#include <stdlib.h>
#include <unistd.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>
#include <X11/keysymdef.h>
#include <X11/keysym.h>
#include <X11/extensions/XTest.h>

/***************************************************************************** 
 * The delay in milliseconds when sending events to the remote display
 ****************************************************************************/
const int DefaultDelay = 10;

/***************************************************************************** 
 * The multiplier used fot scaling coordinates before sending them to the
 * remote display. By default we don't scale at all
 ****************************************************************************/
const float DefaultScale = 1.0;

/***************************************************************************** 
 * Globals...
 ****************************************************************************/
int   Delay = DefaultDelay;
float Scale = DefaultScale;
char * Remote;

/***************************************************************************** 
 * Key used for quitting the application.
 ****************************************************************************/
unsigned int QuitKey;
bool HasQuitKey = false;

/***************************************************************************** 
 * Key used for sending a single key. --Insomnia (Stea Greene)
 ****************************************************************************/
unsigned int SendKey;
bool HasSendKey = false;

/***************************************************************************** 
 * Stream for reading/writing a trace file.
 ****************************************************************************/
ifstream intrace;
ofstream outtrace;
bool writetrace = false;
bool exectrace = false;


/****************************************************************************/
/*! Prints the usage, i.e. how the program is used. Exits the application with
    the passed exit-code.

	\arg const int ExitCode - the exitcode to use for exiting.
*/
/****************************************************************************/
void usage (const int exitCode) {

  // print the usage
  cout << "xremote " << VERSION << endl
       << endl
       << "usage: xremote [options] remote_display" << endl
       << "options: " << endl
       << "  -d  DELAY   delay in milliseconds for events sent to remote display." << endl
	   << "              Default: 10ms." << endl
	   << "  -s  FACTOR  scalefactor for coordinates. Default: 1.0." << endl
	   << "  -k  KEYCODE the keycode for the key used for quitting." << endl
	   << "  -p  KEYCODE send only this keypress then quit." << endl
	   << "  -t  FILE    save trace to file." << endl
	   << "  -x  FILE    execute trace from file." << endl
	   << "  -v          show version. " << endl
	   << "  -h          this help. " << endl
       << endl;

  // we're done
  exit ( EXIT_SUCCESS );
}


/****************************************************************************/
/*! Prints the version of the application and exits.
*/
/****************************************************************************/
void version () {

  // print the version
  cout << "xremote " << VERSION << endl;
 
  // we're done
  exit ( EXIT_SUCCESS );
}


/****************************************************************************/
/*! Opens a trace file for writing to. The file is opened writing and truncated. If the
    opening fails then this function will exit. The global variable trace will be
    modified.

    \arg const char * filename - the file to open for writing.
*/
/****************************************************************************/
void openWriteTraceFile (const char * filename) {

  // try to open the file
  outtrace.open ( filename );

  // did we manage to open it?
  if ( ! outtrace ) {
    // could not open the file
    cout << "xremote: could not open trace file: " << filename << endl;
    exit ( EXIT_FAILURE );
  }
}


/****************************************************************************/
/*! Opens a trace file for reading from. The file is not modified. If the
    opening fails then this function will exit. The global variable trace will be
    modified.

    \arg const char * filename - the file to open for reading.
*/
/****************************************************************************/
void openReadTraceFile (const char * filename) {

  // try to open the file
  intrace.open ( filename );

  // did we manage to open it?
  if ( ! intrace ) {
    // could not open the file
    cout << "xremote: could not open trace file: " << filename << endl;
    exit ( EXIT_FAILURE );
  }
}



/****************************************************************************/
/*! Parses the commandline and stores all data in globals (shudder). Exits
    the application with a failed exitcode if a parameter is illegal. This 
    function is quite ugly and should be rewritten using getopt().

	\arg int argc - number of commandline arguments.
	\arg char * argv[] - vector of the commandline argument strings.
*/
/****************************************************************************/
void parseCommandLine (int argc, char * argv[]) {

  int index = 1;
  
  // check the number of arguments
  if ( argc < 2 ) {
	// oops, too few arguments, go away
	usage ( EXIT_FAILURE );
  }

  // loop through all arguments except the last, which is assumed to be the
  // name of the display
  while ( index < argc ) {
	
	// is this '-v'?
	if ( strcmp (argv[index], "-v" ) == 0 ) {
	  // yep, show version and exit
	  version ();
	}

	// is this '-h'?
	if ( strcmp (argv[index], "-h" ) == 0 ) {
	  // yep, show usage and exit
	  usage ( EXIT_SUCCESS );
	}

	// is this '-d'?
	else if ( strcmp (argv[index], "-d" ) == 0 && index + 1 < argc ) {
	  // yep, and there seems to be a parameter too, interpret it as a
	  // number
	  if ( sscanf ( argv[index + 1], "%d", &Delay ) != 1 ) {
		// oops, not a valid intereger
		cout << "xremote: invalid parameter for '-d'." << endl;
		usage ( EXIT_FAILURE );
	  }
	  
	  index++;
	}

	// is this '-s'?
	else if ( strcmp (argv[index], "-s" ) == 0 && index + 1 < argc ) {
	  // yep, and there seems to be a parameter too, interpret it as a
	  // floating point number
	  if ( sscanf ( argv[index + 1], "%f", &Scale ) != 1 ) {
		// oops, not a valid intereger
		cout << "xremote: invalid parameter for '-s'." << endl;
		usage ( EXIT_FAILURE );
	  }
	  
	  index++;
	}

	// is this '-t'?
	else if ( strcmp (argv[index], "-t" ) == 0 && index + 1 < argc ) {
	  // yep, and there seems to be a parameter too, try to open it as a file
      openWriteTraceFile ( argv[index + 1] );

      // are we already reading a trace? if so this is not legal, as we can't both read
      // and write a trace at the same time
      if ( exectrace ) {
        cout << "xremote: can not both read and write a trace file at the same time." << endl;
        usage ( EXIT_FAILURE );
      }

      // we're tracing now
      writetrace = true;
	  index++;
	}

	// is this '-x'?
	else if ( strcmp (argv[index], "-x" ) == 0 && index + 1 < argc ) {
	  // yep, and there seems to be a parameter too, try to open it as a file for reading
     openReadTraceFile ( argv[index + 1] );

      // are we already writing a trace? if so this is not legal, as we can't both read
      // and write a trace at the same time
      if ( writetrace ) {
        cout << "xremote: can not both read and write a trace file at the same time." << endl;
        usage ( EXIT_FAILURE );
      }

      // we're execing a trace now
      exectrace = true;
	  index++;
	}

    // is this '-k'?
	else if ( strcmp (argv[index], "-k" ) == 0 && index + 1 < argc ) {
	  // yep, and there seems to be a parameter too, interpret it as a
	  // number
	  if ( sscanf ( argv[index + 1], "%d", &QuitKey ) != 1 ) {
		// oops, not a valid integer
		cout << "xremote: invalid parameter for '-k'." << QuitKey << endl;
		usage ( EXIT_FAILURE );
	  }

	  // now we have a key for quitting
	  HasQuitKey = true;
	  index++;
	}
	
    // is this '-p'?  Modeled after '-k' above -- Insomnia (Stea Greene)
	else if ( strcmp (argv[index], "-p" ) == 0 && index + 1 < argc ) {
	  // do we already have the needed key?
	  if ( HasSendKey ) {
		cout << "xremote: can't specify more than one '-p' option." << endl;
		usage ( EXIT_FAILURE );
	  }

	  // parse out the key
	  if ( sscanf ( argv[index + 1], "%d", &SendKey ) != 1 ) {
		// oops, not a valid integer
		cout << "xremote: invalid parameter for '-p'." << SendKey << endl;
		usage ( EXIT_FAILURE );
	  }

	  // now we have a key to send
	  HasSendKey = true;
	  index++;
	}
	
	// is this the last parameter?
	else if ( index == argc - 1 ) {
	  // yep, we assume it's the display, store it
	  Remote = argv [ index ];
	}

	else {
	  // we got this far, the parameter is no good...
	  cout << "xremote: invalid parameter '" << argv[index] << "'." << endl;
	  usage ( EXIT_FAILURE );
	}

	// next value
	index++;
  }
}


/****************************************************************************/
/*! Connects to the local display. Returns the \c Display or \c 0 if
    no display could be obtained.
*/
/****************************************************************************/
Display * localDisplay () {

  // open the display
  Display * D = XOpenDisplay ( 0 );

  // did we get it?
  if ( ! D ) {
	// nope, so show error and abort
	cout << "xremote: could not open display \"" << XDisplayName ( 0 )
		 << "\", aborting." << endl;
	exit ( EXIT_FAILURE );
  }

  // return the display
  return D;
}

/****************************************************************************/
/*! Connects to the desired display. Returns the \c Display or \c 0 if
    no display could be obtained.

	\arg const char * DisplayName - name of the remote display.
*/
/****************************************************************************/
Display * remoteDisplay (const char * displayname) {

  int event, error;
  int major, minor;  

  // open the display
  Display * display = XOpenDisplay ( displayname );

  // did we get it?
  if ( ! display ) {
	// nope, so show error and abort
	cout << "xremote: could not open display \"" << XDisplayName ( displayname )
		 << "\", aborting." << endl;
	exit ( EXIT_FAILURE );
  }

  // does the remote display have the Xtest-extension?
  if ( ! XTestQueryExtension (display, &event, &error, &major, &minor ) ) {
	// nope, extension not supported
	cout << "xremote: XTest extension not supported on server \""
		 << DisplayString(display) << "\"" << endl;

	// close the display and go away
	XCloseDisplay ( display );
	exit ( EXIT_FAILURE );
  }

  // print some information
  cout << "xremote: XTest extension for server \"" << DisplayString(display) << "\" is version "
	   << major << "." << minor << "." << endl << endl;;

  // execute requests even if server is grabbed 
  XTestGrabControl ( display, True ); 

  // sync the server
  XSync ( display, True ); 

  // return the display
  return display;
}


/****************************************************************************/
/*! Function that finds out the key the user wishes to use for quitting the
    application. This must be configurable, as a suitable key is not always
	possible to determine in advance. By letting the user pick a key one that
	does not interfere with the needed applications can be chosen.

	The function grabs the keyboard and waits for a key to be pressed. Returns
	the keycode of the pressed key.

    \arg Display * display - used display.
	\arg int Screen - the used screen.
*/
/****************************************************************************/
int findQuitKey (Display * display, int Screen) { 

  XEvent    event;
  XKeyEvent EKey;
  Window    Target, Root;
  KeySym    keysym;
  int       Error;
  char      Buffer [64];
  char *    keysymstring;
 
  // get the root window and set default target
  Root   = RootWindow ( display, Screen );
  Target = None;

  // grab the keyboard
  Error = XGrabKeyboard ( display, Root, False, GrabModeSync, GrabModeAsync, CurrentTime );

  // did we succeed in grabbing the keyboard?
  if ( Error != GrabSuccess) {
	// nope, abort
	cout << "xremote: could not grab the keyboard, aborting." << endl;
	exit ( EXIT_FAILURE );
  }

  // print a message to the user informing about what's going on
  cout << endl
	   << "Press the key you want to use to end the application. "
	   << "This key can be any key, " << endl
	   << "as long as you don't need it while working with the remote display." << endl
	   << "A good choice is Escape." << endl << endl;

  // discard all X events before we start. This should get rid of the possible 'release'
  // event for the Return key that was generated when this application was started from
  // the command line. If started from within some other application I have no clue what
  // evil this will cause.
  XSync ( display, True );

  // loop until we get the quitting key
  while ( true ) {
    // allow one more event
	XAllowEvents ( display, SyncPointer, CurrentTime);	
    XWindowEvent ( display, Root, KeyPressMask, &event);
	
	// what did we get?
    if ( event.type == KeyPress ) {
	  // a key was pressed, don't loop more and store the key
	  EKey = event.xkey;
      break;
	}
  } 
  
  // now loop until the same key was released
  do {
    // allow one more event
	XAllowEvents ( display, SyncPointer, CurrentTime);	
    XWindowEvent ( display, Root, KeyReleaseMask, &event);
  }
  while ( event.type == KeyRelease );

  // we're done with pointer and keyboard
  XUngrabPointer  ( display, CurrentTime );      
  XUngrabKeyboard ( display, CurrentTime );      

  // perform lookup of the keycode
  XLookupString ( &EKey, Buffer, 64, &keysym, 0 );

  // map the keysym to a string. The string is from <X11/keysymdef.h> and is without the
  // prefix XK_, thus being suitable for saving
  keysymstring = XKeysymToString ( keysym );

  // show the user what was chosen
  cout << "xremote: chosen quit-key: '" << keysymstring << "', keycode: " << EKey.keycode << endl;
  
  // return the found key
  return EKey.keycode;
}


/****************************************************************************/
/*! Scales the passed coordinate with the given saling factor. the factor is
    either given as a commandline argument or it is 1.0.
*/
/****************************************************************************/
int scale (const int Coordinate) {

  // perform the scaling, all in one ugly line
  return (int)( (float)Coordinate * Scale );
}


/****************************************************************************/
/*! Calculates the delta time between the passed time and the last time. If this
    function is called for the first time then 0 is returned.

    \arg const Time time - the new time.
*/
/****************************************************************************/
int calculateDelta (const Time time) {
  static Time last = 0;

  // do we have a last time?
  if ( last == 0 ) {
    // no last time, set this new as the last
    last = time;

    return 0;
  }

  // get the delta
  Time delta = time - last;

  // store new
  last = time;

  return delta;
}




/****************************************************************************/
/*! Sends the key \a Code to the remote display \a RemoteDpy. This function
    sends both a key down and a key up, and the quits. Used if only one
	single keyevent should be sent.

	\arg Display * RemoteDpy - used display.
	\arg const unsigned int Code - keycode of pressed key.
*/
/****************************************************************************/
void sendOnlyOneKey (Display * RemoteDpy, const unsigned int Code) {

  // send the key down
  XTestFakeKeyEvent ( RemoteDpy, Code, True, Delay );

  // flush display
  XFlush ( RemoteDpy );

  // now send the key up
  XTestFakeKeyEvent ( RemoteDpy, Code, False, Delay );

  // flush that too
  XFlush ( RemoteDpy ); 

  // we're done with the entire application
  exit ( EXIT_SUCCESS );
}


/****************************************************************************/
/*! Sends the key \a Code to the remote display \a RemoteDpy. The keycode is
    converted to a \c KeySym on the local display and then reconverted to
	a \c KeyCode on the remote display. Seems to work quite ok, apart from
	something weird with the Alt key.

	\arg XKeyEvent * Event - the local key event
    \arg Display * LocalDpy - used display.
	\arg Display * RemoteDpy - used display.
	\arg const unsigned int Code - keycode of pressed key.
	\arg bool Pressed - true if it is a keypress event, false if a keyrelease.
*/
/****************************************************************************/
void sendKey (XKeyEvent * Event, Display * LocalDpy, Display * RemoteDpy,
			  const unsigned int Code, bool Pressed) {

  KeySym keysym;
  char Buffer [64];
  char * keysymstring;
  KeyCode RemoteKeyCode;
  
  // perform lookup of the keycode
  XLookupString ( Event, Buffer, 64, &keysym, 0 );

  // map the keysym to a string. The string is from <X11/keysymdef.h> and is without the
  // prefix XK_, thus being suitable for saving
  keysymstring = XKeysymToString ( keysym );
    
  // convert it to a keycode on the remote server
  if ( ( RemoteKeyCode = XKeysymToKeycode ( RemoteDpy, keysym ) ) == 0 ) {
  	// no keycode on the remote display for the keysym
  	cout << "xremote: no keycode on remote display found for keysym: " << keysymstring << endl;
  	return;
  }

  // send the event. Check if it's a release or press
  if ( Pressed ) {
    // keypress
    XTestFakeKeyEvent ( RemoteDpy, (unsigned int)RemoteKeyCode, True, Delay );

    // write to the trace file too if we're doing a trace
    if ( writetrace )
      outtrace << "keydown " << calculateDelta ( Event->time ) << " "
            << XKeysymToString ( keysym ) << endl;
  }

  else {
	// keyrelease
	XTestFakeKeyEvent ( RemoteDpy,(unsigned int) RemoteKeyCode, False, Delay );

    // write to the trace file too o if we're doing a trace
    if ( writetrace )
      outtrace << "keyup " << calculateDelta ( Event->time ) << " "
            << XKeysymToString ( keysym ) << endl;
  }
}


/****************************************************************************/
/*! Converts a passed string to a X11 KeyCode and returns it.

    \arg string key - the keycode's string representation.
    \arg Display * display - the display used for the conversion.
*/
/****************************************************************************/
KeyCode stringToKeySym (const string & key, Display * display) {

  // get the keysym for the string
  KeySym keysym = XStringToKeysym ( key.c_str () );

  // now convert the keysym to a keycode and return it
  return XKeysymToKeycode ( display, keysym );
}


/****************************************************************************/
/*! Main loop for playing up a precreated trace file. This function will read all
    lines from the opened trace file and execute the commands one by one.

   \arg Display * LocalDpy - used display.
	\arg Display * RemoteDpy - used display.
	\arg int LocalScreen - the used screen.
	\arg int RemoteScreen - the used screen.
*/
/****************************************************************************/
void playTrace (Display * LocalDpy, int LocalScreen, Display * RemoteDpy, int RemoteScreen) {

  string       command;
  Time         delay;
  unsigned int x, y, button;
  string       key;
  KeyCode      keycode;

  // let the user select a window...
  while ( intrace >> command >> delay ) {
    // first sleep
    usleep ( (unsigned int)delay * 1000 );
    
    // got a command and a delay, what do we have?
    if ( command == "mousemove" ) {
      // read the x and y coordinate
      intrace >> x >> y;

      // send the event. no scaling here, it is already done in the file
      XTestFakeMotionEvent ( RemoteDpy, RemoteScreen, x, y, Delay );
    }

    else if ( command == "mouseup" ) {
      // read the button
      intrace >> button;

      // send the event
      XTestFakeButtonEvent ( RemoteDpy, button, False, Delay );
    }

    else if ( command == "mousedown" ) {
      // read the button
      intrace >> button;

      // send the event
      XTestFakeButtonEvent ( RemoteDpy, button, True, Delay );
    }

    else if ( command == "keydown" ) {
      // read the key string
      intrace >> key;

      // convert the string to a keycode proper for the remote server
      keycode = stringToKeySym ( key, RemoteDpy );

      // send the key event
      XTestFakeKeyEvent ( RemoteDpy, (unsigned int)keycode, True, Delay );
    }

    else if ( command == "keyup" ) {
      // read the key string
      intrace >> key;

      // convert the string to a keycode proper for the remote server
      keycode = stringToKeySym ( key, RemoteDpy );

      // send the key event
      XTestFakeKeyEvent ( RemoteDpy, (unsigned int)keycode, False, Delay );
    }

    else {
      // unknown command
      cout << "xremote: unknown command '" << command << "' in trace file, aborting." << endl;
      return;
    }
    
    // sync the remote server
    XFlush ( RemoteDpy );
  } 
}


/****************************************************************************/
/*! Main event-loop of the application. Loops until a key with the keycode
    \a QuitKey is pressed. Sends all mouse- and key-events to the remote
	display.

    \arg Display * LocalDpy - used display.
	\arg Display * RemoteDpy - used display.
	\arg int LocalScreen - the used screen.
	\arg int RemoteScreen - the used screen.
	\arg unsigned int QuitKey - the key when pressed that quits the eventloop.
*/
/****************************************************************************/
void eventLoop (Display * LocalDpy, int LocalScreen,
				Display * RemoteDpy, int RemoteScreen, unsigned int QuitKey) {

  int          Status1, Status2;
  bool         Loop = true;
  XEvent       Event;
  Window       Root;
  XButtonEvent EButton;
  XMotionEvent EMotion;
  XKeyEvent    EKey;
  
  // get the root window and set default target
  Root = RootWindow ( LocalDpy, LocalScreen );
  
  // grab the pointer 
  Status1 = XGrabPointer ( LocalDpy, Root, False,
						   PointerMotionMask|ButtonPressMask|ButtonReleaseMask,
						   GrabModeSync, GrabModeAsync, Root, None, CurrentTime );

  // grab the keyboard
  Status2 = XGrabKeyboard ( LocalDpy, Root, False, GrabModeSync, GrabModeAsync, CurrentTime );
  
  // did we succeed in grabbing the pointer?
  if ( Status1 != GrabSuccess) {
	// nope, abort
	cout << "xremote: could not grab the local mouse, aborting." << endl;
	exit ( EXIT_FAILURE );
  }

  // did we succeed in grabbing the keyboard?
  if ( Status2 != GrabSuccess) {
	// nope, abort
	cout << "xremote: could not grab the local keyboard, aborting." << endl;
	exit ( EXIT_FAILURE );
  }

  // let the user select a window...
  while ( Loop ) {
    // allow one more event
	XAllowEvents ( LocalDpy, SyncPointer, CurrentTime);	

	// get an event matching the specified mask
	XWindowEvent ( LocalDpy, Root,
				   KeyPressMask|KeyReleaseMask|PointerMotionMask|ButtonPressMask|ButtonReleaseMask,
				   &Event);
	
	// what did we get?
    switch (Event.type) {
    case ButtonPress:
	  // button pressed, create event
	  EButton = Event.xbutton;
	  XTestFakeButtonEvent ( RemoteDpy, EButton.button, True, Delay );

      // log the fact too if we're doing a trace
      if ( writetrace )
        outtrace << "mousedown " << calculateDelta ( EButton.time ) << " " << EButton.button << endl;
      break;

    case ButtonRelease:
	  // button released, create event
	  EButton = Event.xbutton;
	  XTestFakeButtonEvent ( RemoteDpy, EButton.button, False, Delay );

      // log the fact too if we're doing a trace
      if ( writetrace )
        outtrace << "mouseup " << calculateDelta ( EButton.time ) << " " << EButton.button << endl;
	  break;

	case MotionNotify:
	  // motion-event, create event
	  EMotion = Event.xmotion;
	  XTestFakeMotionEvent ( RemoteDpy, RemoteScreen ,
							 scale ( EMotion.x ), scale ( EMotion.y ), Delay ); 

      // log the fact too if we're doing a trace
      if ( writetrace )
        outtrace << "mousemove " << calculateDelta ( EMotion.time ) << " " << scale (EMotion.x)
              << " " << scale ( EMotion.y ) << endl;
	  break;

	case KeyRelease:
	  // a key was released
	  EKey = Event.xkey;
	  sendKey  ( &EKey, LocalDpy, RemoteDpy, EKey.keycode, false );
	  break;
	  
	case KeyPress:
	  // a key was pressed, don't loop more
	  EKey = Event.xkey;
	  
	  // should we stop looping, i.e. did the user press the quitkey?
	  if ( EKey.keycode == QuitKey ) {
		// yep, no more loops
		Loop = false;
	  }
	  else {
		// send the keycode to the remote server
		sendKey  ( &EKey, LocalDpy, RemoteDpy, EKey.keycode, true );
	  }
	  break;
	}

	// sync the remote server
	XFlush ( RemoteDpy );
  } 

  // we're done with pointer and keyboard
  XUngrabPointer  ( LocalDpy, CurrentTime );      
  XUngrabKeyboard ( LocalDpy, CurrentTime );      
}


/****************************************************************************/
/*! Main function of the application. It expects no commandline arguments.

    \arg int argc - number of commandline arguments.
	\arg char * argv[] - vector of the commandline argument strings.
*/
/****************************************************************************/
int main (int argc, char * argv[]) {

  // parse commandline arguments
  parseCommandLine ( argc, argv );
  
  // open the local display
  Display * LocalDpy  = localDisplay ();
  Display * RemoteDpy = remoteDisplay ( Remote );

  // get the screens too
  int LocalScreen  = DefaultScreen ( LocalDpy );
  int RemoteScreen = DefaultScreen ( RemoteDpy );

  // send sole key and quit if desired.
  if ( HasSendKey ) {
	// send one key and quit
	sendOnlyOneKey ( RemoteDpy, (unsigned int)SendKey );
  }

  else if ( exectrace ) {
    // we're executing a trace file, so go do it
    playTrace ( LocalDpy, LocalScreen, RemoteDpy, RemoteScreen );
  }
  
  else {
    // we're doing more than just sending a single key or executing a trace. Do we have a
    // registered quitting key? If one was supplied as a commandline argument we use that
    // key 
    if ( ! HasQuitKey ) {
      // nope, so find the key that quits the application
      QuitKey = findQuitKey ( LocalDpy, LocalScreen );
    }

    // start the main event loop
    eventLoop ( LocalDpy, LocalScreen, RemoteDpy, RemoteScreen, QuitKey );
  }

  // discard and even flush all events on the remote display
  XTestDiscard ( RemoteDpy );
  XFlush ( RemoteDpy ); 

  // we're done with the display
  XCloseDisplay ( RemoteDpy );
  XCloseDisplay ( LocalDpy );

  cout << "xremote: pointer and keyboard released. " << endl;
  
  // go away
  exit ( EXIT_SUCCESS );
}

/*  
 * Local Variables:
 * mode: auto-fill
 * fill-column: 90
 * End:
 */


